// $Header: /users/cvsleni/cvs/GA/galib/ga/gaallelef.h,v 1.2 1999/09/15 14:13:14 adam Exp $
/* ----------------------------------------------------------------------------
   allele.h
   mbwall 21mar95
   Copyright (c) 1995 Massachusetts Institute of Technology
   all rights reserved

   DESCRIPTION:
   Here we define a class of alleles.  An allele is a possible value for a gene
   and an allele set is a list of possible values (I use 'set' because it doesn't
   imply the specific implementation of the container class).

   MODIFICATIONS:
---------------------------------------------------------------------------- */
#ifndef _ga_allele_h_
#define _ga_allele_h_
//#include <string.h>
#include <ga/gaconfig.h>
#include <ga/garandom.h>

#ifndef NO_STREAMS
#include <iostream>
#endif

#ifndef T
#define T float
#endif


class GAAllele {
public:
  enum Type {ENUMERATED=1, BOUNDED=2, DISCRETIZED};
  enum BoundType {NONE, INCLUSIVE, EXCLUSIVE};
};


/* ----------------------------------------------------------------------------
  This object contains a set of alleles for use with a similarly typed genome.
  This object can be used with any type that has operator= defined for it.  If
you use the remove member then you must have operator== defined for it.
  This should be implemented as a derivative of the Array class?  But I don't
want that overhead at this point.  Also, behaviour is not the same.
  The allele set uses the envelope/message structure.  The core allele object
is a reference-counted structure that contains all of the guts of an allele
set.  The outer shell, the allele set itself, is what users actually see.  It
defines the interface.  With this setup you can create a single allele set then
each genome does not have to make its own copy.  And we don't have to worry
about an allele set going out of scope then leaving genomes hanging (a problem
if we just used a pointer to a single alleleset with no reference counting).
  You can link an allele set to another so that they share the same core.  Use
the 'link' member function (this is typically used within the GAlib to reduce 
the number of actual alleleset instances when cloning populations of genomes).
  There is no way to 'resize' an allele set.  You must add to it or remove
elements from it.
  The base template class assumes that the objects in the allele set are 
complex, i.e. it is not OK to do a bit-copy of each object.  We should do 
specializations for non-complex objects (or perhaps a separate class that does
bit-copies rather than logical-copies?)
  When you clone an allele set, the new one has its own core.
  Why didn't I do this as a couple of objects (enumerated set, bounded set, 
discretized set, etc)?  I wanted to be able to have an array that uses sets of
many different types.  I suppose the allele set should be completely 
polymorphic like the rest of the GAlib objects, but for now we'll do it as
a single object with multiple personalities (and a state).
  There is no error checking.  You should check the type before you try to
call any of the member functions.  In particular, if you try to get the
bounds on an enumerated set of one element, it will break.

*** should the assignment operator check to be sure that no allele is 
    duplicated, or is it OK to have duplicate alleles in a set?  For now we
    allow duplicates (via either the add or assignemnt ops).
---------------------------------------------------------------------------- */

class GAAlleleSetCore {
public:
  GAAlleleSetCore();
  GAAlleleSetCore(unsigned int n, const T array [],const float w=1.0);
  GAAlleleSetCore(const T& lower, const T& upper, 
		  GAAllele::BoundType lb=GAAllele::INCLUSIVE, 
		  GAAllele::BoundType ub=GAAllele::INCLUSIVE,const float w=1.0);
  GAAlleleSetCore(const T& lower, const T& upper, const T& increment,
		  GAAllele::BoundType lb=GAAllele::INCLUSIVE, 
		  GAAllele::BoundType ub=GAAllele::INCLUSIVE,const float w=1.0);
  GAAlleleSetCore(const GAAlleleSetCore&);
  virtual ~GAAlleleSetCore();
  GAAlleleSetCore& operator=(const GAAlleleSetCore&);

  GAAllele::Type type;		// is this an ennumerated or bounded set?
  GAAllele::BoundType lowerb, upperb; // what kind of limit is the bound?
  unsigned int cnt;		// how many objects are using us?
  unsigned int csz;		// how big are the chunks to allocate?
  unsigned int sz;		// number we have
  unsigned int SZ;		// how many have we allocated?
  T * a;
  float weight;			/* how important in crossover etc. */
};


// mods for symantec C++ on macOS.  thanks to RJG 15jul96
#if defined(THINK_C) || __powerc
 int operator==(const T &, const T &);
 int operator!=(const T &, const T &);
#endif


class GAAlleleSet {
public:
  GAAlleleSet() : core(0) {}
  GAAlleleSet(unsigned int n, const T a[],const float w=1.0) : 
    core(new GAAlleleSetCore(n,a,w)) {}
  GAAlleleSet(const T& lower, const T& upper, 
	      GAAllele::BoundType lb=GAAllele::INCLUSIVE, 
	      GAAllele::BoundType ub=GAAllele::INCLUSIVE,const float w=1.0) :
    core(new GAAlleleSetCore(lower, upper, lb, ub,w)) {}
  GAAlleleSet(const T& lower, const T& upper, const T& increment,
	      GAAllele::BoundType lb=GAAllele::INCLUSIVE, 
	      GAAllele::BoundType ub=GAAllele::INCLUSIVE,const float w=1.0) :
    core(new GAAlleleSetCore(lower, upper, increment, lb, ub,w)) {}
  GAAlleleSet(const GAAlleleSet& set) : 
    core(new GAAlleleSetCore(*(set.core))) {}
  virtual ~GAAlleleSet(){
    if(core != 0){
      core->cnt -= 1;
      if(core->cnt==0) delete core;
    }
  }
  GAAlleleSet & operator=(const GAAlleleSet& set){
    if(this == &set) return *this;
    if(core) *core = *set.core;
    else core = new GAAlleleSetCore(*(set.core));
    return *this;
  } 
  GAAlleleSet * clone() const {return new GAAlleleSet(*this);}
  void link(GAAlleleSet& set);
  void unlink();
  int size() const {return core->sz;}    // only meaninful for enumerated sets
  int add(const T&);		         // only for enumerated sets
  int remove(const T&);			 // only for enumerated sets
  int remove(unsigned int);		 // only for enumerated sets
  T allele() const;		         // ok for any type of set
  T allele(unsigned int i) const;        // only for enumerated sets
  T lower() const {return core->a[0];}   // only for bounded sets
  T upper() const {return core->a[1];}
  T inc() const {return core->a[2];}
  float weight() const {return core->weight;}
  float weight(const float w) const {return (core->weight=w);}
  GAAllele::BoundType lowerBoundType() const {return core->lowerb;}
  GAAllele::BoundType upperBoundType() const {return core->upperb;}
  GAAllele::Type type() const {return core->type;}

#ifndef NO_STREAMS
  int read(istream &);
  int write(ostream & os) const;
#endif

#if defined(THINK_C) || __powerc
  friend operator==(const GAAlleleSet &, const GAAlleleSet &);
  friend operator!=(const GAAlleleSet &, const GAAlleleSet &);
#else
  friend int operator==(const GAAlleleSet &, const GAAlleleSet &);
  friend int operator!=(const GAAlleleSet &, const GAAlleleSet &);
#endif

protected:
  GAAlleleSetCore *core;
};


 
class GAAlleleSetArray {
public:
  GAAlleleSetArray();
  GAAlleleSetArray(const GAAlleleSet& s);
  GAAlleleSetArray(const GAAlleleSetArray&);
  virtual ~GAAlleleSetArray();
  GAAlleleSetArray& operator=(const GAAlleleSetArray&);

  int size() const {return sz;}
  const GAAlleleSet& set(unsigned int i) const {return *(aset[i]);}
  int add(const GAAlleleSet& s);
  int add(unsigned int n, const T a[],const float w=1.0);
  int add(const T& lower, const T& upper,
	  GAAllele::BoundType lb=GAAllele::INCLUSIVE,
	  GAAllele::BoundType ub=GAAllele::INCLUSIVE,const float w=1.0);
  int add(const T& lower, const T& upper, const T& increment,
	  GAAllele::BoundType lb=GAAllele::INCLUSIVE, 
	  GAAllele::BoundType ub=GAAllele::INCLUSIVE,const float w=1.0);
  int remove(unsigned int);

#ifndef NO_STREAMS
  int read(istream &);
  int write(ostream & os) const;
#endif

protected:
  unsigned int csz;
  unsigned int sz;
  unsigned int SZ;
  GAAlleleSet **aset;

};


#ifndef NO_STREAMS
inline ostream& 
operator<< (ostream& os, const GAAlleleSet & arg)
{ arg.write(os); return os; }
inline istream& 
operator>> (istream& is, GAAlleleSet & arg)
{ arg.read(is); return is; }

inline ostream& 
operator<< (ostream& os, const GAAlleleSetArray & arg)
{ arg.write(os); return os; }
inline istream& 
operator>> (istream& is, GAAlleleSetArray & arg)
{ arg.read(is); return is; }
#endif



#undef T 



#ifdef BORLAND_INST
#include <ga/GAAllele.C>
#endif

#endif
